from typing import List, Dict, Optional
import numpy as np
import torch
from tqdm import tqdm
from metric.claim_extractor.claim_extractor import ClaimExtractor
from metric.coreference_resolution.coreference_resolution import CoreferenceResolution
from metric.nli.nli_aligner import NLIAligner
from metric.utils.utils import split_into_paragraphs, split_into_sentences_batched


class FENICE:
    def __init__(
        self,
        use_coref: bool = False,
        num_sent_per_paragraph: int = 5,
        sliding_paragraphs=True,
        sliding_stride: int = 1,
        doc_level_nli=True,
        paragraph_level_nli=True,
        claim_extractor_batch_size: int = 256,
        coreference_batch_size: int = 1,
        nli_batch_size: int = 256,
        nli_max_length: int = 1024,
    ) -> None:
        self.num_sent_per_paragraph = num_sent_per_paragraph
        self.claim_extractor_batch_size = claim_extractor_batch_size
        self.coreference_batch_size = coreference_batch_size
        self.nli_batch_size = nli_batch_size
        self.sliding_paragraphs = sliding_paragraphs
        self.sliding_stride = sliding_stride
        self.sentences_cache = {}
        self.coref_clusters_cache = {}
        self.claims_cache = {}
        self.alignments_cache = {}
        self.use_coref = use_coref
        self.doc_level_nli = doc_level_nli
        self.paragraph_level_nli = paragraph_level_nli
        self.coref_model = (
            CoreferenceResolution(load_model=False) if use_coref else None
        )
        self.nli_max_length = nli_max_length
        self.device = "cuda" if torch.cuda.is_available() else "cpu"

    def _score(self, sample_id: int, document: str, summary: str):
        doc_id = self.get_id(sample_id, document)
        sentences_offsets = self.sentences_cache[doc_id]
        sentences = [s[0] for s in sentences_offsets]
        offsets = [(s[1], s[2]) for s in sentences_offsets]
        # paragraphs are sliding windows of {num_sent_per_paragraph} sentences
        paragraphs = split_into_paragraphs(
            sentences,
            self.num_sent_per_paragraph,
            sliding_paragraphs=self.sliding_paragraphs,
            sliding_stride=self.sliding_stride,
        )
        # claim extraction
        summary_id = self.get_id(sample_id, summary)
        summary_claims = self.claims_cache.get(summary_id, [summary])
        alignments = []
        for claim_id, claim in enumerate(summary_claims):
            # sentence-level alignment
            sentence_level_alignment = self.get_alignment(
                premises=sentences,
                hypothesis=claim,
                sample_id=sample_id,
                hypothesis_id=claim_id,
            )
            coref_alignment = None
            if self.use_coref:
                coref_clusters = self.coref_clusters_cache[doc_id]
                # modified versions of {aligned_sentence} obtained through coreference resolution
                coref_premises = self.coref_model.get_coref_versions(
                    sentence=sentence_level_alignment["source_passage"],
                    text=document,
                    sentences=sentences,
                    offsets=offsets,
                    clusters=coref_clusters,
                )
                if coref_premises:
                    coref_alignment = self.get_alignment(
                        sample_id=sample_id,
                        hypothesis_id=claim_id,
                        hypothesis=claim,
                        premises=coref_premises,
                        alignment_prefix="coref",
                    )
            paragraph_level_alignment = None
            if len(paragraphs) > 1 and self.paragraph_level_nli:
                paragraph_level_alignment = self.get_alignment(
                    premises=paragraphs,
                    hypothesis=claim,
                    sample_id=sample_id,
                    hypothesis_id=claim_id,
                    alignment_prefix="par",
                )
            doc_level_alignment = None
            if self.doc_level_nli and len(paragraphs) > 1:
                doc_level_alignment = self.get_alignment(
                    hypothesis=claim,
                    premises=[document],
                    sample_id=sample_id,
                    hypothesis_id=claim_id,
                    alignment_prefix="doc",
                )
                doc_level_alignment["source_passage"] = "DOCUMENT"
            sample_alignments = [
                sentence_level_alignment,
                coref_alignment,
                paragraph_level_alignment,
                doc_level_alignment,
            ]
            alignment = self.max_alignment(sample_alignments)
            alignments.append(alignment)
        score = np.mean([al["score"] for al in alignments])
        return {"score": score, "alignments": alignments}

    def max_alignment(self, sample_alignments):
        sample_alignments = [s for s in sample_alignments if s is not None]
        alignment = max(sample_alignments, key=lambda x: x["score"])
        return alignment

    def get_alignment(
        self,
        premises,
        hypothesis,
        sample_id,
        hypothesis_id,
        alignment_prefix: Optional[str] = None,
    ):
        alignments_ids = []
        for prem_id in range(len(premises)):
            pair_id = self.get_alignment_id(
                sample_id=sample_id,
                premise_id=prem_id,
                hypothesis_id=hypothesis_id,
                premise=premises[prem_id],
                alignment_prefix=alignment_prefix,
            )
            alignments_ids.append(pair_id)
        loaded_alignment = self.load_alignment(
            alignments_ids=alignments_ids, premises=premises, hypothesis=hypothesis
        )
        if loaded_alignment is None:
            all_pairs = [(x, hypothesis) for x in premises]
            self.cache_alignment(alignments_ids, all_pairs)
            alignment, _ = self.load_alignment(
                alignments_ids=alignments_ids, premises=premises, hypothesis=hypothesis
            )
        else:
            alignment, _ = loaded_alignment
        return alignment

    def score_batch(self, batch: List[Dict[str, str]]) -> List[Dict]:
        documents = [el["document"] for el in batch]
        summaries = [el["summary"] for el in batch]
        self.cache(documents, summaries)
        predictions = []
        for sample_id, (doc, summary) in tqdm(
            enumerate(zip(documents, summaries)),
            total=len(documents),
            desc="Computing FENICE...",
        ):
            predictions.append(self._score(sample_id, doc, summary))

        del self.nli_aligner
        torch.cuda.empty_cache()

        return predictions

    def cache(self, documents, summaries):
        self.cache_sentences(documents)
        if self.use_coref:
            self.cache_coref(documents)
        self.cache_claims(summaries)
        self.cache_alignments(documents, summaries)

    def cache_sentences(self, documents):
        all_sentences = split_into_sentences_batched(
            documents, batch_size=512, return_offsets=True
        )
        for i, sentences in enumerate(all_sentences):
            id = self.get_id(i, documents[i])
            self.sentences_cache[id] = sentences

    def get_docs_to_process(self, documents, cache):
        ids = [doc_id for doc_id in range(len(documents))]
        ids_to_process = [
            id for id in ids if self.get_id(id, documents[id]) not in cache
        ]
        doc_to_process = [documents[i] for i in ids_to_process]
        return doc_to_process, ids_to_process

    def get_id(self, sample_id: int, text: str, k_chars: int = 100):
        id = f"{sample_id}{text[:k_chars]}"
        return id

    # cache claim extraction outputs
    def cache_claims(self, summaries):
        claim_extractor = ClaimExtractor(
            batch_size=self.claim_extractor_batch_size, device=self.device
        )
        claims_predictions = claim_extractor.process_batch(summaries)
        for summ_id, claims in enumerate(claims_predictions):
            id = self.get_id(summ_id, summaries[summ_id])
            self.claims_cache[id] = claims
        del claim_extractor
        torch.cuda.empty_cache()

    # cache coreference resolution outputs
    def cache_coref(self, documents):
        coref_model = CoreferenceResolution(
            batch_size=self.coreference_batch_size, device=self.device
        )
        all_clusters = coref_model.get_clusters_batch(documents)
        for doc_id, clusters in enumerate(all_clusters):
            id = self.get_id(doc_id, documents[doc_id])
            self.coref_clusters_cache[id] = clusters
        del coref_model.model
        torch.cuda.empty_cache()

    def cache_alignments(self, documents: List[str], summaries: List[str]):
        alignments_ids, all_pairs = [], []
        alignment_ids_doc, all_pairs_doc = [], []
        self.nli_aligner = NLIAligner(
            batch_size=self.nli_batch_size,
            device=self.device,
            max_length=self.nli_max_length,
        )
        for sample_id in range(len(summaries)):
            summary_id = self.get_id(sample_id, summaries[sample_id])
            document_id = self.get_id(sample_id, documents[sample_id])
            claims = self.claims_cache[summary_id]
            sentences = [s[0] for s in self.sentences_cache[document_id]]
            paragraphs = split_into_paragraphs(
                sentences,
                self.num_sent_per_paragraph,
                sliding_paragraphs=self.sliding_paragraphs,
                sliding_stride=self.sliding_stride,
            )
            alignments_ids, all_pairs = self.compute_nli_pairs(
                alignments_ids, all_pairs, claims, sample_id, sentences
            )
            if self.paragraph_level_nli:
                alignments_ids, all_pairs = self.compute_nli_pairs(
                    alignments_ids,
                    all_pairs,
                    claims,
                    sample_id,
                    paragraphs,
                    prefix="par",
                )
            if self.doc_level_nli:
                alignment_ids_doc, all_pairs_doc = self.compute_nli_pairs(
                    alignment_ids_doc,
                    all_pairs_doc,
                    claims,
                    sample_id,
                    [documents[sample_id]],
                    prefix="doc",
                )
        self.cache_alignment(alignments_ids, all_pairs)
        self.nli_aligner.batch_size = 1
        self.nli_aligner.max_length = 4096
        self.cache_alignment(alignment_ids_doc, all_pairs_doc)
        self.nli_aligner.batch_size = self.nli_batch_size
        self.nli_max_length = self.nli_max_length

    def compute_nli_pairs(
        self, alignments_ids, all_pairs, claims, sample_id, premises, prefix=None
    ):
        for premise_id, premise in enumerate(premises):
            for claim_id, claim in enumerate(claims):
                alignment_id = self.get_alignment_id(
                    sample_id, premise_id, claim_id, premise, prefix
                )
                alignments_ids.append(alignment_id)
                all_pairs.append((premise, claim))
        return alignments_ids, all_pairs

    def get_alignment_id(
        self,
        sample_id: int,
        premise_id: int,
        hypothesis_id: int,
        premise: str,
        alignment_prefix: Optional[str] = None,
        k_chars: int = 100,
    ):

        id = f"{sample_id}-{premise_id}-{hypothesis_id}-{premise[:k_chars]}"
        if alignment_prefix is not None:
            id = f"{alignment_prefix}{id}"
        return id

    def cache_alignment(
        self, alignments_ids, all_pairs, disable_prog_bar: bool = False
    ):
        probabilities = self.nli_aligner.process_batch(
            all_pairs, disable_prog_bar=disable_prog_bar
        )
        for i, id in enumerate(alignments_ids):
            ent, contr, neut = (
                probabilities[0][i],
                probabilities[1][i],
                probabilities[2][i],
            )
            self.alignments_cache[id] = (ent, contr, neut)

    def load_alignment(
        self, alignments_ids: List[str], premises: List[str], hypothesis: str
    ):
        if all([k in self.alignments_cache for k in alignments_ids]):
            scores = [self.alignments_cache[key] for key in alignments_ids]
            alignment = None
            max_score = -np.inf
            for i, (ent, contr, neut) in enumerate(scores):
                ent, contr, neut = ent.item(), contr.item(), neut.item()
                align_score = ent - neut
                if align_score > max_score:
                    max_score = align_score
                    alignment = (premises[i], [ent, contr, neut])
            return (
                {
                    "score": max_score,
                    "summary_claim": hypothesis,
                    "source_passage": alignment[0],
                },
                scores,
            )
        else:
            return None



import hashlib
import pickle
from typing import List, Tuple, Union
import spacy
import sys
import os
from spacy.cli import download
from tqdm import tqdm

try:
    nlp = spacy.load("en_core_web_sm")
except Exception as e:
    os.environ["SPACY_WARNING_IGNORE"] = "true"  # Prevent interactive prompts
    download("en_core_web_sm")
    nlp = spacy.load("en_core_web_sm")


def flatten(two_d_list: List[List]) -> List:
    """
    Flattens a 2D Python list into a 1D list using list comprehension.
    """
    return [item for sublist in two_d_list for item in sublist]


def chunks(lst: List, n: int) -> List[List]:
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i : i + n]


def sliding_chunks(lst: List, n: int, sliding_stride: int = 1) -> List[List]:
    """Yield sliding windows of n-sized chunks from lst."""
    for i in range(len(lst) - n + 1):
        if i % sliding_stride == 0:
            yield lst[i : i + n]


def distinct(input_list: List) -> List:
    seen = set()
    return [x for x in input_list if not (x in seen or seen.add(x))]


def split_into_sentences(
    text: str, return_offsets: bool = False
) -> Union[List[Tuple[str, str, str]], List[str]]:
    # Process the text with spaCy
    doc = nlp(text)
    return get_sentences(doc, return_offsets)


def get_sentences(doc, return_offsets):
    # Initialize a list to store sentences and their offsets
    sentences_with_offsets = []
    # Iterate over the sentences in the processed text
    for sent in doc.sents:
        # Get the sentence text and character offsets
        sentence_text = sent.text
        start_offset = sent.start_char
        end_offset = sent.end_char

        # Add the sentence text and offsets to the list as a tuple
        sentences_with_offsets.append((sentence_text, start_offset, end_offset))
    if return_offsets:
        return sentences_with_offsets
    else:
        # Return only the sentences
        return [sentence[0] for sentence in sentences_with_offsets]


def split_into_sentences_batched(
    texts: List[str], return_offsets: bool = False, batch_size=32
) -> List[List[Tuple[str, str, str]]]:
    # Process the text with spaCy
    batches = list(chunks(texts, batch_size))
    sentences = []
    for b in tqdm(
        batches, total=len(batches), desc="splitting document batches into sentences"
    ):
        docs = nlp.pipe(b, disable=["attribute_ruler", "lemmatizer", "ner"])
        for doc in docs:
            doc_sentences = get_sentences(doc, return_offsets=return_offsets)
            sentences.append(doc_sentences)
    return sentences


def split_into_paragraphs(
    sentences: List[str],
    num_sent_per_paragraph: int,
    sliding_paragraphs=True,
    sliding_stride: int = 1,
) -> List[str]:
    if len(sentences) < num_sent_per_paragraph:
        return [" ".join(sentences)]
    paragraphs = (
        list(sliding_chunks(sentences, num_sent_per_paragraph, sliding_stride))
        if sliding_paragraphs
        else list(chunks(sentences, num_sent_per_paragraph))
    )
    for i, par in enumerate(paragraphs):
        paragraphs[i] = " ".join(par)
    return paragraphs


def hash_text(text: str) -> str:
    return hashlib.md5(text.encode("utf-8")).hexdigest()


def load_pickle(path: str):
    # Open the file in binary read mode
    with open(path, "rb") as file:
        # Load the object from the file
        return pickle.load(file)


def dump_pickle(path: str, data):
    # Open the file in binary write mode
    with open(path, "wb") as file:
        # Dump the object to the file
        pickle.dump(data, file)



from typing import List, Tuple
from fastcoref import FCoref
from tqdm import tqdm

from metric.utils.utils import chunks


class CoreferenceResolution:
    def __init__(
        self,
        input_max_toks: int = 10_000,
        pronouns_path: str = "data/pronouns.txt",
        batch_size: int = 1,
        load_model: bool = True,
        device: str = "cuda:0",
    ):
        self.input_max_toks = input_max_toks
        self.model = None
        if load_model:
            self.model = FCoref(device=device, enable_progress_bar=False)
        self.batch_size = batch_size
        with open(pronouns_path) as f:
            self.pronouns = f.read().splitlines()

    def get_clusters(self, document: str) -> List[List[Tuple[int, int]]]:
        # we truncate the document to a maximum length before executing coreference resolution
        # document = self.truncate_document(document)
        preds = self.model.predict(texts=[document])
        clusters = []
        if preds:
            clusters = preds[0].get_clusters(as_strings=False)
        return clusters

    # def truncate_document(self, document):
    #     return " ".join(document.split()[: self.input_max_toks])

    def get_clusters_batch(
        self, documents: List[str]
    ) -> List[List[List[Tuple[int, int]]]]:
        batched_documents = list(chunks(documents, n=self.batch_size))
        clusters = []
        for batch in tqdm(
            batched_documents,
            desc="Executing coreference resolution",
            total=len(batched_documents),
        ):
            # for i, doc in enumerate(batch):
            #     batch[i] = self.truncate_document(doc)
            batch_preds = self.model.predict(texts=batch)
            clusters.extend(
                [pred.get_clusters(as_strings=False) for pred in batch_preds]
            )
        return clusters

    # returns modified versions of {sentence} obtained through coreference resolution
    def get_coref_versions(
        self,
        sentence: str,
        text: str,
        sentences: List[str],
        offsets: List[Tuple[int, int]],
        clusters: List[List[Tuple[int, int]]],
    ):
        max_aligned_idx = [
            i for i, sent in enumerate(sentences) if sent.strip() == sentence.strip()
        ]
        if max_aligned_idx:
            max_aligned_idx = max_aligned_idx[0]
        else:
            return []
        document_sent_start, document_sent_end = offsets[max_aligned_idx]
        document_mentions = self.get_mentions(
            clusters, document_sent_start, document_sent_end
        )
        candidates = []
        for cluster in document_mentions:
            target_offsets = [
                (off[0], off[1])
                for off in cluster
                if document_sent_start <= off[0] and off[1] < document_sent_end
            ]
            target_words = [text[off[0] : off[1]] for off in target_offsets]
            relative_offsets = [
                (off[0] - document_sent_start, off[1] - document_sent_start)
                for off in target_offsets
            ]
            other_mentions = [
                text[ment[0] : ment[1]]
                for ment in cluster
                if text[ment[0] : ment[1]] not in target_words
            ]
            other_mentions = set(
                [m for m in other_mentions if m.lower() not in self.pronouns]
            )
            for target_word, target_offset in zip(target_words, relative_offsets):
                candidate = sentence[: target_offset[0]]
                for ment in other_mentions:
                    edited_candidate = sentence[target_offset[0] :].replace(
                        target_word, ment, 1
                    )
                    edited_candidate = candidate + edited_candidate
                    candidates.append(edited_candidate)
        return list(set(candidates))

    def get_mentions(
        self, clusters: List[List[Tuple[int, int]]], sent_start: int, sent_end: int
    ):
        return [
            cluster
            for cluster in clusters
            if any(
                [
                    (sent_start <= offset[0] and offset[1] < sent_end)
                    for offset in cluster
                ]
            )
        ]



from typing import List

from tqdm import tqdm
from transformers import T5ForConditionalGeneration, T5Tokenizer
from metric.utils.utils import chunks, split_into_sentences, distinct


class ClaimExtractor:
    def __init__(
        self,
        model_name: str = "Babelscape/t5-base-summarization-claim-extractor",
        device: str = "cuda:0",
        batch_size: int = 70,
    ):
        self.device = device
        # load model from HF
        self.model = T5ForConditionalGeneration.from_pretrained(model_name).to(device)
        self.tokenizer = T5Tokenizer.from_pretrained(model_name)
        self.batch_size = batch_size

    def process_batch(self, batch: List[str]) -> List[List[str]]:
        predictions = []
        batches = list(chunks(batch, self.batch_size))
        for b in tqdm(batches, desc="Extracting claims..."):
            tok_input = self.tokenizer.batch_encode_plus(
                b, return_tensors="pt", padding=True
            ).to(self.device)
            claims = self.model.generate(**tok_input)
            claims = self.tokenizer.batch_decode(claims, skip_special_tokens=True)
            claims = [split_into_sentences(c) for c in claims]
            claims = [distinct(c) for c in claims]
            predictions.extend(claims)
        return predictions



from setuptools import setup, find_packages

# Call the function during setup
setup(
    name="FENICE",
    version="0.1.14",
    author="Alessandro Scirè",
    author_email="scire@diag.uniroma1.it",
    description="This package contains the code to execute FENICE, a factuality-oriented metric for summarization",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/Babelscape/FENICE",
    license="CC BY-NC-SA 4.0",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=[
        "numpy",
        "torch",
        "tqdm",
        "spacy==3.7.4",
        "fastcoref==2.1.6",
        "transformers~=4.38.2",
        "sentencepiece==0.2.0",
        "scikit-learn==1.5.0",
    ],
    entry_points={
        "console_scripts": [
            # If you have command line scripts, list them here.
            # For example: 'my-script=my_package.module:function',
        ]
    },
)



from tqdm import tqdm
from transformers import AutoTokenizer, AutoModelForSequenceClassification
import torch
from metric.utils.utils import chunks


class NLIAligner:
    def __init__(
        self,
        model_name: str = "MoritzLaurer/DeBERTa-v3-large-mnli-fever-anli-ling-wanli",
        batch_size: int = 32,
        device: str = "cuda:0",
        max_length: int = 512,
        **kwargs,
    ):
        self.model_name = model_name
        self.tokenizer = AutoTokenizer.from_pretrained(model_name)

        self.device = device
        self.model = AutoModelForSequenceClassification.from_pretrained(model_name).to(
            device
        )
        self.batch_size = batch_size
        self.max_length = max_length

    def score(self, premises, hypothesis, lower=False, disable_prog_bar: bool = False):
        if isinstance(hypothesis, str):
            hypothesis = [hypothesis] * len(premises)
        if lower:
            premises = [p.lower() for p in premises]
            hypothesis = [h.lower() for h in hypothesis]
        prem_hyp_pairs = list(zip(premises, hypothesis))
        return self.process_batch(prem_hyp_pairs, disable_prog_bar=disable_prog_bar)

    def process_batch(self, prem_hyp_pairs, disable_prog_bar=False):
        batches = list(chunks(prem_hyp_pairs, self.batch_size))
        all_entailment_probs = []
        all_contradiction_probs = []
        all_neutral_probs = []
        for batch in tqdm(
            batches,
            desc="Computing alignments...",
            total=len(batches),
            disable=disable_prog_bar,
        ):
            entailment_dists = self.score_sample(batch)
            all_entailment_probs.extend([d[0] for d in entailment_dists])
            all_neutral_probs.extend([d[1] for d in entailment_dists])
            all_contradiction_probs.extend(d[2] for d in entailment_dists)
        return all_entailment_probs, all_contradiction_probs, all_neutral_probs

    def score_sample(self, batch):
        tokenized_input_seq_pairs = self.tokenizer.batch_encode_plus(
            batch,
            return_token_type_ids=True,
            padding=True,
            return_tensors="pt",
            truncation=True,
            max_length=self.max_length,
        ).to(self.device)
        input_ids = tokenized_input_seq_pairs["input_ids"]
        token_type_ids = tokenized_input_seq_pairs["token_type_ids"]
        attention_mask = tokenized_input_seq_pairs["attention_mask"]
        with torch.no_grad():
            outputs = self.model(
                input_ids,
                attention_mask=attention_mask,
                token_type_ids=token_type_ids,
                labels=None,
            )

        entailment_dists = torch.softmax(outputs.logits, dim=1)
        return entailment_dists
